CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

OBJ
  c : "clibsd"

PUB start
  c.start
  main

{
 ############################################################################
 # This program is used to test the basic functions of the SD file system.
 # It implements simple versions of the cat, rm, ls, echo, cd, pwd, mkdir and
 # rmdir commands plus the <, > and >> file redirection operators.
 # The program starts up the file driver and then prompts for a command.
 #
 # Written by Dave Hein
 # Copyright (c) 2011,2012 Parallax, Inc.
 # MIT Licensed
 ############################################################################
}

CON
  ATTR_READ_ONLY = 1
  ATTR_ARCHIVE   = $20
  ATTR_DIRECTORY = $10

  ' typedef struct dirent
  attr       = 11
  filesize_0 = 28
  d_name     = 32

VAR
  long stdinfile
  long stdoutfile
  
{ Print help information }
PUB Help
  c.printf0(string("Commands are help, cat, rm, ls, ll, echo, cd, pwd, mkdir, rmdir and chmod\n"))

PUB Cd(argc, argv)
  if (argc < 2)
    return
  if (c.chdir(long[argv][1]))
    c.perror(long[argv][1])

PUB Move(argc, argv)
  if (argc <> 3)
    c.printf0(string("usage: mv name1 name2\n"))
    return
  if (c.rename(long[argv][1], long[argv][2]))
    c.perror(0)

PUB Pwd(argc, argv) | buffer[16], ptr
  ptr := c.getcwd(@buffer, 64)
  if (not ptr)
    c.perror(0)
  else
    c.fprintf1(stdoutfile, string("%s\n"), ptr)

PUB Mkdir(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.mkdir(long[argv][i], 0))
      c.perror(long[argv][i])
    i++

PUB Rmdir(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.rmdir(long[argv][i]))
      c.perror(long[argv][i])
    i++
    
{ This routine implements the file cat function }
PUB Cat(argc, argv) | i, num, infile, buffer[10]
  i := 0
  repeat while (i < argc)
    if (i == 0)
      if (argc == 1 or stdinfile <> c.getstdin)
        infile := stdinfile
      else
        i++
        next
    else
      infile := c.fopen(long[argv][i], string("r"))
      if (infile == 0)
        c.perror(long[argv][i])
        i++
        next
    if (infile == c.getstdin)
      repeat while (c.gets(@buffer))
        if (byte[@buffer][0] == 4)
          quit
        c.fprintf1(stdoutfile, string("%s\n"), @buffer)
    else
      repeat while ((num := c.fread(@buffer, 1, 40, infile)))
        c.fwrite(@buffer, 1, num, stdoutfile)
    if (i)
      c.fclose(infile)
    i++
  c.fflush(c.getstdout)
  
{ This routine deletes the files specified by the command line arguments }
PUB Remove(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (c.remove(long[argv][i]))
      c.perror(long[argv][i])
    i++
    
{ This routine echos the command line arguments }
PUB Echo(argc, argv) | i
  i := 1
  repeat while (i < argc)
    if (i <> argc - 1)
      c.fprintf1(stdoutfile, string("%s "), long[argv][i])
    else
      c.fprintf1(stdoutfile, string("%s\n"), long[argv][i])
    i++
    
{ This routine lists the root directory or any subdirectories specified
   in the command line arguments.  If the "-l" option is specified, it
   will print the file attributes and size.  Otherwise, it will just
   print the file names.  }
PUB List(argc, argv) | i, j, fname, count, filesize, longflag, path, drwx[2], column, prevlen, dirp, entry
  count := 0
  longflag := 0
  ' Check flags
  j := 1
  repeat while (j < argc)
    if (byte[long[argv][j]][0] == "-")
      if (not c.strcmp(long[argv][j], string("-l")))
        longflag := 1
      else
        c.printf1(string("Unknown option \022%s\022\n"), long[argv][j])
    else
      count++
    j++
  ' List directories
  j := 1
  repeat while (j < argc or count == 0)
    if (count == 0)
      count--
      path := string("./")
    elseif (byte[long[argv][j]][0] == "-")
      j++
      next
    else
      path := long[argv][j]
    if (count => 2)
      c.fprintf1(stdoutfile, string("\n%s:\n"), path)
    dirp := c.opendir(path)
    if (not dirp)
      c.perror(path)
      j++
      next
    column := 0
    prevlen := 14
    repeat while (entry := c.readdir(dirp))
      if (byte[entry + d_name][0] == ".")
        next
      fname := entry + d_name
      filesize := long[entry + filesize_0]
      c.strcpy(@drwx, string("-rw-"))
      if (byte[entry + attr] & ATTR_READ_ONLY)
        byte[@drwx][2] := "-"
      if (byte[entry + attr] & ATTR_ARCHIVE)
        byte[@drwx][3] := "x"
      if (byte[entry + attr] & ATTR_DIRECTORY)
        byte[@drwx][0] := "d"
      if (longflag)
        c.fprintf3(stdoutfile, string("%s %8d %s\n"), @drwx, filesize, fname)
      elseif (++column == 5)
        i := prevlen
        repeat while (i < 14)
          c.fprintf0(stdoutfile, string(" "))
          i++
        c.fprintf1(stdoutfile, string("%s\n"), fname)
        column := 0
        prevlen := 14
      else
        i := prevlen
        repeat while (i < 14)
          c.fprintf0(stdoutfile, string(" "))
          i++
        prevlen := strsize(fname)
        c.fprintf1(stdoutfile, string("%s"), fname)
    c.closedir(dirp)
    if (not longflag and column)
      c.fprintf0(stdoutfile, string("\n"))
    j++
    
PUB Chmod(argc, argv) | ptr, fname, value, setmask, setflag, bitpos, buf[8]
  if argc <> 3
    ChmodUsage
    return
  ptr := long[argv][1]
  fname := long[argv][2]
  if c.stat(fname, @buf)
    c.perror(fname)
    return
  setmask := byte[@buf][attr]
  'Convert from FAT to UNIX format
  setmask := ((setmask&16)>>1) | 4 | (((setmask&1)^1)<<1) | ((setmask&32)>>5)
  if byte[ptr] => "0" AND byte[ptr] =< "7"
    c.sscanf1(ptr, string("%o"), @value)
    ifnot value & !7
      setmask := (setmask & 8) | value
    else
      ChmodUsage
      return
    'Convert from UNIX to FAT format
    setmask := ((setmask&8)<<1) | ((setmask&1)<<5) | (((setmask&2)>>1)^1)
    c.chmod(fname, setmask)
    return
  if byte[ptr] <> "+" AND byte[ptr] <> "-"
    ChmodUsage
    return
  setflag := 1
  repeat while byte[ptr]
    value := byte[ptr++]
    if value == "+"
      setflag := 1
      next
    elseif value == "-"
      setflag := 0
      next
    elseif value == "x"
      bitpos := 1
    elseif value == "w"
      bitpos := 2
    elseif value == "r"
      bitpos := 4
    else
      ChmodUsage
      return
    if setflag
      setmask |= bitpos
    else
      setmask &= (bitpos ^ %1111)
  'Convert from UNIX to FAT format
  setmask := ((setmask&8)<<1) | ((setmask&1)<<5) | (((setmask&2)>>1)^1)
  c.chmod(fname, setmask)

PUB ChmodUsage
  c.printf0(string("usage: chmod mode file\n"))
  c.printf0(string("  mode is either an octal digit or\n"))
  c.printf0(string("  contains the characters +-rwx\n"))

{ This routine returns a pointer to the first character that doesn't
   match val. }
PUB SkipChar(ptr, val)
  repeat while (byte[ptr])
    if (byte[ptr] <> val)
      quit
    ptr++
  return ptr
  
{ This routine returns a pointer to the first character that matches val. }
PUB FindChar(ptr, val)
  repeat while (byte[ptr])
    if (byte[ptr] == val)
      quit
    ptr++
  return ptr
  
{ This routine extracts tokens from a string that are separated by one or
   more spaces.  It returns the number of tokens found. }
PUB tokenize(ptr, tokens) | num
  num := 0
  repeat while (byte[ptr])
    ptr := SkipChar(ptr, " ")
    if (byte[ptr] == 0)
      quit
    if (byte[ptr][0] == ">")
      ptr++
      if (byte[ptr][0] == ">")
        long[tokens][num++] := string(">>")
        ptr++
      else
        long[tokens][num++] := string(">")
      next
    if (byte[ptr][0] == "<")
      ptr++
      long[tokens][num++] := string("<")
      next
    long[tokens][num++] := ptr
    ptr := FindChar(ptr, " ")
    if (byte[ptr])
      byte[ptr++] := 0
  return num
  
{ This routine searches the list of tokens for the redirection operators
   and opens the files for input, output or append depending on the 
   operator. }
PUB CheckRedirection(tokens, num) | i, j
  i := 0
  repeat while (i < num - 1)
    if (not c.strcmp(long[tokens][i], string(">")))
      stdoutfile := c.fopen(long[tokens][i + 1], string("w"))
      if (not stdoutfile)
        c.perror(long[tokens][i + 1])
        stdoutfile := c.getstdout
        return 0
    elseif (not c.strcmp(long[tokens][i], string(">>")))
      stdoutfile := c.fopen(long[tokens][i + 1], string("a"))
      if (not stdoutfile)
        c.perror(long[tokens][i + 1])
        stdoutfile := c.getstdout
        return 0
    elseif (not c.strcmp(long[tokens][i], string("<")))
      stdinfile := c.fopen(long[tokens][i + 1], string("r"))
      if (not stdinfile)
        c.perror(long[tokens][i + 1])
        stdinfile := c.getstdin
        return 0
    else
      i++
      next
    j := i + 2
    repeat while (j < num)
      long[tokens][j - 2] := long[tokens][j]
      j++
    i--
    num -= 2
    i++
  return num
  
{ This routine closes files that were open for redirection }
PUB CloseRedirection
  if (stdinfile <> c.getstdin)
    c.fclose(stdinfile)
    stdinfile := c.getstdin
  if (stdoutfile <> c.getstdout)
    c.fclose(stdoutfile)
    stdoutfile := c.getstdout

PUB mount | mountErr
  c.printf0(string("Load and mount SD: "))
  ' Important: This code assumes you're using a C3 card.
  ' If you're using different hardware, make sure you
  ' change the following initialization to match your card!
  mountErr := c.mount_explicit(10, 11, 9, 25)
  'mountErr := c.mount_explicit(0, 1, 2, 3)
  if mountErr
    c.printf1(string("Mount error: %d\n"), mountErr)
    abort 1
  c.printf0(string("done.\n\n"))
  
{ The program starts the file system.  It then loops reading commands
   and calling the appropriate routine to process it. }
PUB main | num, tokens[20], buffer[20]
  stdinfile := c.getstdin
  stdoutfile := c.getstdout
  ' Wait for the serial terminal to start
  waitcnt(CNT + CLKFREQ*3)
  mount
  Help
  repeat while (1)
    c.printf0(string("\n> "))
    c.fflush(c.getstdout)
    c.gets(@buffer)
    num := tokenize(@buffer, @tokens)
    num := CheckRedirection(@tokens, num)
    if (num == 0)
      next
    if (not c.strcmp(tokens[0], string("help")))
      Help
    elseif (not c.strcmp(tokens[0], string("cat")))
      Cat(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("ls")))
      List(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("ll")))
      tokens[num++] := string("-l")
      List(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("rm")))
      Remove(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("echo")))
      Echo(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("cd")))
      Cd(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("pwd")))
      Pwd(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("mkdir")))
      Mkdir(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("rmdir")))
      Rmdir(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("mv")))
      Move(num, @tokens)
    elseif (not c.strcmp(tokens[0], string("chmod")))
      Chmod(num, @tokens)
    else
      c.printf0(string("Invalid command\n"))
      Help
    CloseRedirection
  return 0
{
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
 (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:
The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
}